"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GcodeParser = void 0;
// Adapted from https://github.com/cncjs/gcode-parser/blob/master/src/index.js
var GcodeParser = /** @class */ (function () {
    function GcodeParser() {
    }
    GcodeParser.prototype.parseLine = function (line, options) {
        if (options === void 0) { options = {}; }
        var fullOptions = __assign({ flatten: !!options.flatten, noParseLine: !!options.noParseLine }, (options || {}));
        var result = {
            line: line,
        };
        if (fullOptions.noParseLine) {
            return result;
        }
        result.words = [];
        var ln; // Line number
        var cs; // Checksum
        var words = this.stripComments(line).match(this.constructor.re) || [];
        for (var i = 0; i < words.length; ++i) {
            var word = words[i];
            var letter = word[0].toUpperCase();
            var argument = word.slice(1);
            // Parse % commands for bCNC and CNCjs
            // - %wait Wait until the planner queue is empty
            if (letter === '%') {
                result.cmds = (result.cmds || []).concat(line.trim());
                continue;
            }
            // Parse JSON commands for TinyG and g2core
            if (letter === '{') {
                result.cmds = (result.cmds || []).concat(line.trim());
                continue;
            }
            // Parse $ commands for Grbl
            // - $C Check gcode mode
            // - $H Run homing cycle
            if (letter === '$') {
                result.cmds = (result.cmds || []).concat("".concat(letter).concat(argument));
                continue;
            }
            // N: Line number
            if (letter === 'N' && typeof ln === 'undefined') {
                // Line (block) number in program
                ln = Number(argument);
                continue;
            }
            // *: Checksum
            if (letter === '*' && typeof cs === 'undefined') {
                cs = Number(argument);
                continue;
            }
            var value = Number(argument);
            if (Number.isNaN(value)) {
                value = argument;
            }
            if (fullOptions.flatten) {
                result.words.push(letter + value);
            }
            else {
                result.words.push([letter, value]);
            }
        }
        // Line number
        (typeof (ln) !== 'undefined') && (result.ln = ln);
        var _a = this.getStringMessage(line), command = _a[0], stringMessage = _a[1];
        if (command && stringMessage) {
            if (fullOptions.flatten) {
                result.words = [command, stringMessage];
            }
            else {
                var commandLetter = command[0];
                var commandNumber = command.slice(1);
                if (!Number.isNaN(Number(commandNumber))) {
                    commandNumber = Number(commandNumber);
                }
                var parameterName = this.constructor
                    .STRING_MESSAGE_PARAMETER_NAME[commandNumber];
                result.words = [
                    [commandLetter, commandNumber],
                    [parameterName, stringMessage],
                ];
            }
            return result;
        }
        // Checksum
        (typeof (cs) !== 'undefined') && (result.cs = cs);
        var actualCs = this.computeChecksum(line);
        if (result.cs && (actualCs !== result.cs)) {
            result.err = true; // checksum failed
            result.actualCs = actualCs;
        }
        return result;
    };
    GcodeParser.prototype.getStringMessage = function (line) {
        var matchStringMessage = Array.from(line.matchAll(this.constructor.reStringMessage))[0];
        if (!matchStringMessage) {
            return [null, null];
        }
        var command = "M" + matchStringMessage[2].trim();
        var message = matchStringMessage[3]
            .trim()
            .replace(/(^|[^\\]|((^|[^\\])\\\\)+);.*/, '$1')
            .replaceAll(/\\([^\\])/g, '$1')
            .trim();
        return [command, message];
    };
    // http://reprap.org/wiki/G-code#Special_fields
    // The checksum "cs" for a GCode string "cmd" (including its line number) is computed
    // by exor-ing the bytes in the string up to and not including the * character.
    GcodeParser.prototype.computeChecksum = function (s) {
        s = s || '';
        if (s.lastIndexOf('*') >= 0) {
            s = s.slice(0, s.lastIndexOf('*'));
        }
        var cs = 0;
        for (var i = 0; i < s.length; ++i) {
            var c = s[i].charCodeAt(0);
            cs ^= c;
        }
        return cs;
    };
    // http://linuxcnc.org/docs/html/gcode/overview.html#gcode:comments
    // Comments can be embedded in a line using parentheses () or for the remainder of a lineusing a semi-colon. The semi-colon is not treated as the start of a comment when enclosed in parentheses.
    GcodeParser.prototype.stripComments = function (line) {
        return line
            .replace(this.constructor.re1, '')
            .replace(this.constructor.re2, '')
            .replace(this.constructor.re3, '');
    };
    GcodeParser.re = /(%.*)|({.*)|(\$\$|\$[a-zA-Z0-9#]*)|([a-zA-Z][0-9+\-.]+)|(\*[0-9]+)/igm;
    // Some commands have a string message, which is not parsed as normally
    GcodeParser.reStringMessage = /\s*([Nn]\s*[0-9+\-.]+)?\s*[Mm]\s*(16|23|28|30|33|117|118|928|81[0-9])([^0-9+\-.].*)/igm;
    GcodeParser.STRING_MESSAGE_PARAMETER_NAME = {
        16: 'string',
        23: 'filename',
        28: 'filename',
        30: 'filename',
        33: 'path',
        117: 'string',
        118: 'string',
        928: 'filename',
        810: 'command',
        811: 'command',
        812: 'command',
        813: 'command',
        814: 'command',
        815: 'command',
        816: 'command',
        817: 'command',
        818: 'command',
        819: 'command',
    };
    GcodeParser.re1 = new RegExp(/\s*\([^)]*\)/g); // Remove anything inside the parentheses
    GcodeParser.re2 = new RegExp(/\s*;.*/g); // Remove anything after a semi-colon to the end of the line, including preceding spaces
    GcodeParser.re3 = new RegExp(/\s+/g);
    return GcodeParser;
}());
exports.GcodeParser = GcodeParser;
//# sourceMappingURL=GcodeParser.js.map